#include <Arduino.h>

// include the SPI library:
#include <SPI.h>

#include <InstrumentedObjectHelpers.h> //contains debug output tool
using namespace helpers;               // contains the verbose levels from InstrumentedObjectHelpers.h

#include <BufferSupport.h> //data buffers
// #include <SerialSupport.h>  //Serial object

const float versionNumber = 1.2;    // version displayed on boot
bool outputFloat = false;          // true = ADC converted to float, false = ADC in 16 bit integer
const uint16_t serialTimeOut = 1000; // timeout in ms
const long serialBaudrate = 115200;
const byte numChars = 32; //maximum length of messages parsed on receive.
static uint16_t logFreqSamples = 200; // give samplecounter and summary every n samples
uint8_t downSampler = 16;        // average n number of samples before storing. 1 = no downsampling, 2 = average two samples
const uint8_t maxDownSampler = 200;    // max number of samples to average
const uint16_t minSampleRate = 1; // Lowest Acceptable Sample Rate (Hz)
const uint32_t maxSampleRate = 100000; // Highest Acceptable Sample Rate (Hz)

const uint8_t channelCountSize = 1;                          // always fixed at 1 byte
const unsigned char channelCount[channelCountSize] = {0x01}; // number of active channels send to the plot, for now this is ADC. Should be 0x02 if Batt is also send.

const int spiClockDivider = 42;  // ADC SPI speed
SPISettings spiSettingsA(4000000, MSBFIRST, SPI_MODE1);
const int chipSelectADCPin = 10; // ADC SPI
const u_int8_t MOSIpin = 11;     // ADC SPI //not used, only one way communication
const u_int8_t MISOpin = 12;     // ADC SPI
const u_int8_t SCKpin = 13;      // ADC SPI

byte sensorBuffer[16];
byte rawSensorBuffer[16];
byte sensorStatusBuffer[8];

#define _TASK_TIMECRITICAL       // allows to check timing overrun etc.
#define _TASK_THREAD_SAFE        // allows for multi threading, not used at the moment
#define _TASK_MICRO_RES          // delays etc all in µs instead of ms
#define _TASK_SCHEDULING_OPTIONS // Support for multiple scheduling options
//#define CHECK_TIMING             // enables limits on overrun and jitter on the sensor readout task
#include <TaskScheduler.h>       //https://github.com/arkhipenko/TaskScheduler/
#include <elapsedMillis.h>


// the task durations define the sample rate of the tasks:
long aquisitionTaskDuration = 125;      // µs per cycle   //for aquisition of sensor data
const long communicationTaskDuration = 10000; // µs per cycle   //of data transmission to client via Blueooth
const uint8_t jitterMax = 10;                 // after this number of missed samples, the device is stopped.
int jitterCounter = 0;                        // amount of times the sample momement was more than one sample time off.

// Callback methods prototypes used by the task scheduler
void aquisitionTask_Callback();
void communicationTask_Callback();

// Tasks
Task aquisitionTask(aquisitionTaskDuration, TASK_FOREVER, &aquisitionTask_Callback);
Task communicationTask(communicationTaskDuration, TASK_FOREVER, &communicationTask_Callback);

Scheduler runner; // tasks are added to the scheduler in the setup() sequence and the scheduler is executed in main() loop.

long ADCresult;
long ADCresultArray[maxDownSampler];
float ADCvolts;
byte *bADCvolts[4]; // 4 byte float for uScope format
float battVolts;
byte *bBattvolts[4]; // 4 byte float for uScope format
elapsedMicros elapsedMicrosTime; //timer for aquisition
elapsedMillis elapsedMillisTime; //timer for receiving commands

namespace buffers
{
  // Definition of the data format
  // header (3 bytes): 7f ff bf
  const unsigned char startByteA = 0x7F; //  bytes placed in front of every sample set
  const unsigned char startByteB = 0xFF; //  bytes placed in front of every sample set
  const unsigned char startByteC = 0xBF; //  bytes placed in front of every sample set

  const uint8_t startBytesSize = 3;
  const unsigned char startBytes[startBytesSize] = {startByteA, startByteB, startByteC};

  const int communicationBufferSize = 10240; // size of the two buffers used for communication towards receiver. Due to a bug, all buffers currently have this size, and only the first few positions are used in most cases.

  // initialize buffer names, they will be defined in the setup();

  BufferSupport bufferMillis; // microtime (4 bytes, signed integer): 00 01 01 01   big-e
  BufferSupport bufferADC;    // floats (4 bytes each)
                              // BufferSupport bufferBatt;   // floats (4 bytes each)
  // BufferSupport bufferCRC;          // CRC-16 data buffer, 2x2 byte
  BufferSupport bufferSampleSet; // complete buffer for a single sample set
  BufferSupport btBufferA;       // first "large" buffer used to send the data to the receiver
  BufferSupport btBufferB;       // second "large" buffer used to send the data to the receiver
}
using namespace buffers;


//read messages received, this call is blocking for as long as new data continues to arrive
//default start and end markers are < & >, other can be defines via arguments.
//characters between the start and end marker are parsed as commands. For example "foo<start>foo"  is returned as "start"
//maxReadTime is reserved for future use, to implement a timeout function if new characters keep comming in.
//based on https://forum.arduino.cc/t/serial-input-basics-updated/382007
char *receiveSerialMessage(const char startMarker = '<', const char endMarker = '>', const int maxReadTime = 100)
{
    static char receivedChars[numChars]; // an array to store the received data

    static boolean recvInProgress = false;
    boolean newData = false;
    static byte index = 0;
    char charactersRead;
    elapsedMillisTime = 0;

    //TODO: cancel while loop if new data continues to arrive
    while (Serial.available() > 0 && newData == false && elapsedMillisTime < maxReadTime) //read until no new data is available, or maxReadTime is reached
    {
        charactersRead = Serial.read();

        if (recvInProgress == true)
        {
            if (charactersRead != endMarker)
            {
                receivedChars[index] = charactersRead;
                index++;
                if (index >= numChars)
                {
                    index = numChars - 1; //TODO: comment FM: does not seem logical to do?

                    helpers::printToLog("Received more characters via serial than the max array length!  " + String(receivedChars), verboseLevels::error);
                }
            }
            else
            {
                receivedChars[index] = '\0'; // terminate the string
                recvInProgress = false;
                index = 0;
                newData = true;
            }
        }

        else if (charactersRead == startMarker)
        {
            recvInProgress = true;
        }
    }
    if (newData == true) //print the parsed command for debug purposes.
    {
        helpers::printToLog("Received via Serial: " + String(receivedChars), verboseLevels::info);
    }

    return receivedChars;
}


uint32_t sampleCounter = 0;
uint32_t sampleCounterOld = 0;
uint32_t millisOld = 0;
uint8_t millisWraps = 0;                 // count the number of times millis() reached 2^32.
uint8_t downSampleCounter = 0;          // count up each aquisition cycle


/*
this taks reads the sensor data at a fixed interval. Based upon the realtime example of library TaskScheduler (https://github.com/arkhipenko/TaskScheduler)
this task will be started and stopped based upon the <start> / <stop> commands received via Bluetooth.
 */
void aquisitionTask_Callback()
{
#if defined(CHECK_TIMING) // jitter detection for the interal scheduler
  if (aquisitionTask.getOverrun() != aquisitionTaskDuration)
  {
    if (abs(aquisitionTask.getStartDelay()) > aquisitionTaskDuration) // more than one sample jitter
    {
      helpers::printToLog("too much overrun = " + String(aquisitionTask.getOverrun()) + ", start delayed by " + String(aquisitionTask.getStartDelay()), verboseLevels::extraTiming);
      jitterCounter = jitterCounter + 1; // count the number of times a full sample is missed.
    }
    else
    {
      jitterCounter = 0; // reset the counter if a sample is made within the normal sample time.
    }
  }

  if (jitterCounter > jitterMax)
  {
    helpers::printToLog("too much overrun = " + String(jitterMax) + " times a sample is missed sequentially.", verboseLevels::error);
    // only warn, never stop    // aquisitionTask.disable();
    // helpers::printToLog("Stopped aquisition.", verboseLevels::error, PIN_LED_R);
  }
#endif // !Check timing
  // helpers::printToLog("sample counter: " + String(sampleCounter), verboseLevels::info);
  bool succesRead = true; // TODO: implement Sensor readout check

  // Read ADC

  // A falling edge on CS initiates a conversion and the data transfer. After the fifth DCLOCK falling edge, DOUT is enabled and forced
  // low. The data bits are then clocked, MSB first, by subsequent DCLOCK falling edges. The data is valid on both DCLOCK
  // edges. Although the rising edge can be used to capture the data, a digital host also using the DCLOCK falling edge allows a
  // faster reading rate, provided it has an acceptable hold time.
  SPI.beginTransaction(spiSettingsA);
  digitalWrite(chipSelectADCPin, LOW);   // select the chip
  //delayMicroseconds(2);
  byte adcByteHigh = SPI.transfer(0x00); // transfer 0x00 to the device, store byte received in adcByteHigh
  //delayMicroseconds(2);
  byte adcByteMid = SPI.transfer(0x00);  // transfer 0x00 to the device, store byte received in adcByteMid
  //delayMicroseconds(2);
  byte adcByteLow = SPI.transfer(0x00);  // transfer 0x00 to the device, store byte received in adcByteLow
  //delayMicroseconds(2);
  digitalWrite(chipSelectADCPin, HIGH);  // deselect the chip
  SPI.endTransaction();

  // combine the two parts into one 18-bit number:
  //ADCresult = ((adcByteHigh << 14) | (adcByteMid << 6) | (adcByteLow >> 2));
  //ADCresult = ((adcByteHigh << 16) | (adcByteMid << 8) | (adcByteLow >>0));
  ADCresult = (adcByteHigh & 3) << 14 | adcByteMid << 6 | adcByteLow >> 2; // & 3 part masks the two highest bits of adcByteHigh which should be 0 anyways.
  ADCresultArray[downSampleCounter] = ADCresult;
  
  //downSampler = 4 gives: 0 1 2 3
  downSampleCounter = downSampleCounter +1; //increase counter

  if (downSampleCounter >= downSampler)  // downsampler counter is equal or higher to desired number of downsample times.
  {
    downSampleCounter = 0;    // reset downsample counter
    sampleCounter = sampleCounter + 1;  // sample counter is not reset after start / stop sampling
    if (sampleCounter > (16777216) - 1) // sample counter wrapping at 2^24-1
    {
      helpers::printToLog("sample counter wrap at = " + String(sampleCounter), verboseLevels::info);
      sampleCounter = 0;
      sampleCounterOld = 0;
    }

      
    if (elapsedMicrosTime > (2147483648)) // won't fit in signed INT32    2147483648 = 2^31
    {
        elapsedMicrosTime = 0;  //reset timer 
    }
    uint32_t currentTime = elapsedMicrosTime;
    



    // convert current time to bytes  (32 bit)  //TODO: this is unsigned, uScope expects singed int for time (why!?)
    bufferMillis.buffer[3] = (uint8_t)(currentTime) & 0x80; // LSB   //0x80 instead of 0xFF seems to fix the negative time in uScope
    bufferMillis.buffer[2] = (uint8_t)(currentTime >> 8) & 0xFF;
    bufferMillis.buffer[1] = (uint8_t)(currentTime >> 16) & 0xFF;
    bufferMillis.buffer[0] = (uint8_t)(currentTime >> 24) & 0xFF; // MSB

    bufferSampleSet.clearBuffer();                                                   // clear single sample data buffer
    memcpy(bufferSampleSet.buffer, startBytes, startBytesSize);                      // insert startByte
    memcpy(bufferSampleSet.buffer + startBytesSize, channelCount, channelCountSize); // append channel count  1 byte

    memcpy(bufferSampleSet.buffer + startBytesSize + channelCountSize, bufferMillis.buffer, bufferMillis.getBufferSize()); // append millisecond counter datasample counter

    bufferADC.clearBuffer(); // clear ECG read buffer

    //average the ADC readings according to downsample amount
    ADCresult = 0; //clear ADCresult, which is also used above

    //calculate average of the ADC readings
    
    for (int i = 0; i < downSampler; i++)
    {
    ADCresult += ADCresultArray[i];
    }
    ADCresult = ADCresult / downSampler; //compensate for downsampling
    
    
    if(outputFloat){
    // convert to float voltage
      ADCvolts = (5.0 * ((float)ADCresult / 65535)) ; //convert to volts and convert tp float
      ADCvolts = ADCvolts - 2.5; //offset to compensate for the single supply ADC
    }
    else
    {
       ADCvolts = (float)ADCresult ; //cast to float without conversion into volts
    }

    // convert to bytes
    byte *bADCvolts = (byte *)&ADCvolts;

    // copy to buffer  //TODO: implement range copying
    for (int i = 0; i < bufferADC.getBufferSize(); i++)
    {
      bufferADC.buffer[i] = bADCvolts[i];
    }

    // append ADC data, starting pointer after the time bytes
    memcpy(bufferSampleSet.buffer + startBytesSize + channelCountSize + bufferMillis.getBufferSize(), bufferADC.buffer, bufferADC.getBufferSize());

    // TODO: implement battery readout
    // battVolts = 5.0 * ((float)1 / 65535); // TODO bogus value
    // // convert to bytes
    // byte *bBattvolts = (byte *)&battVolts;

    // bufferBatt.clearBuffer();

    // // copy to buffer  //TODO: implement range copying
    // for (int i = 0; i < bufferBatt.getBufferSize(); i++)
    // {
    //   bufferBatt.buffer[i] = bBattvolts[i];
    // }
    // // append Batt data, starting pointer after the time bytes
    // memcpy(bufferSampleSet.buffer + startBytesSize + channelCountSize + bufferMillis.getBufferSize() + bufferADC.getBufferSize(), bufferBatt.buffer, bufferBatt.getBufferSize());

    // the "bufferSampleSet" is done, all samples of a single sample set are present
    //  the code below will append this "bufferSampleSet" to one of the two btBuffers based upon their lock status. Before writing the btBuffer will be locked, and will not be available for sending. In this way race conditions are prevented.
    if (succesRead) // TODO: this only checks if the Sensor read was successfull, might be useful to check for ADC data as well.
    {
      if (btBufferA.isBufferLocked())
      {
        if (btBufferB.isBufferLocked())
        {
          helpers::printToLog("both write buffers locked!", verboseLevels::error);
        }
        else
        {
          // buffer B not locked, write to buffer B
          memcpy(btBufferB.buffer + btBufferB.getPointerPosition(), bufferSampleSet.buffer, bufferSampleSet.getBufferSize()); // copy read buffer into send buffer
          bufferSampleSet.clearBuffer();                                                                                      // clear read buffer
          btBufferB.setPointerPosition(btBufferB.getPointerPosition() + bufferSampleSet.getBufferSize());                     // move pointer for next write to send buffer
          if (btBufferB.getPointerPosition() > btBufferB.getBufferSize() - bufferSampleSet.getBufferSize())                   // TODO: should this be checked before moving the pointer?
          {
            helpers::printToLog("Buffer B overflow!, dropped Buffer B", verboseLevels::error);
            btBufferB.clearBuffer(); // clear buffer and reset pointer
          }
        }
      }
      else
      {
        // buffer A not locked, write to buffer A
        memcpy(btBufferA.buffer + btBufferA.getPointerPosition(), bufferSampleSet.buffer, bufferSampleSet.getBufferSize()); // copy read buffer into send buffer
        bufferSampleSet.clearBuffer();                                                                                      // clear read buffer
        btBufferA.setPointerPosition(btBufferA.getPointerPosition() + bufferSampleSet.getBufferSize());                     // move pointer for next write to send buffer
        if (btBufferA.getPointerPosition() > btBufferA.getBufferSize() - bufferSampleSet.getBufferSize())                   // TODO: should this be checked before moving the pointer?
        {
          helpers::printToLog("Buffer A overflow!, dropped Buffer A", verboseLevels::error);
          btBufferA.clearBuffer(); // clear buffer and reset pointer
        }
      }
    }
    else
    {
      helpers::printToLog("Sensor read failed at sample counter: " + String(sampleCounter), verboseLevels::error);
    }

    // log every n samples to show progress, toggle led when message is given

    if (sampleCounter - sampleCounterOld > logFreqSamples - 1)
    {
      sampleCounterOld = sampleCounter;

      float millisDelta = millis() - millisOld;
      millisOld = millis();
      float rateHz = 1000 / (millisDelta / logFreqSamples);

      helpers::printToLog("sample counter: " + String(sampleCounter) + ", delta: " + (String)millisDelta + " ms, rate: " + (String)rateHz + " ADC (V) " + (String)ADCvolts, verboseLevels::info);
      // Serial.println("sample counter: " + String(sampleCounter) + ", delta: " + (String)millisDelta + " ms, rate: " + (String)rateHz + " | ");
    }

  } // end downsampler counter check
} // end aquisitionTask

// this task will check if one of the btBuffers contains sufficient data and is unlocked. It will then send it to the Bluetooth receiver.
// this task is executed at regular intervals, and is executed in a different thread as the aquisition task.
// this task will be active right from the boot, and will never be stopped.
void communicationTask_Callback()
{
  //read data from Serial

   // characters between the start and end marker are parsed as commands. For example "foo<start>foo"  is returned as "start"
    const char startMarker = '<';
    const char endMarker = '>';
    const int maxReadTime = 100; //ms
    char *receivedChars = receiveSerialMessage(startMarker, endMarker, maxReadTime);

    if (strcmp(receivedChars, "start") == 0)
    {
      if (!aquisitionTask.isEnabled())
      {
        memset(receivedChars, 0, sizeof(receivedChars) ); // clean command buffer
        jitterCounter = 0; // reset the counter for a clean start
        aquisitionTask.enable();
        helpers::printToLog("Start command received, enabled aquisitionTask (start sampling)", verboseLevels::info);
      }
    }
    else if (strcmp(receivedChars, "stop") == 0)
    {
      if (aquisitionTask.isEnabled())
      {
        memset(receivedChars, 0, sizeof(receivedChars) ); // clean command buffer
        aquisitionTask.disable();
        helpers::printToLog("Stop command received: disabled aquisitionTask (stop sampling).", verboseLevels::info);
      }
    }
    else if (strncmp(receivedChars, "SetSampleRate", 13) == 0) // command starting with "SetSampleRate", should always be followed by 5 integers, for example "SetSampleRate01000" for 1000 Hz
    {
      if (aquisitionTask.isEnabled())
      {
        helpers::printToLog("Received Set Sample Rate command, but aquisition is active!", verboseLevels::error);
      }
      else
      {
          char argument[6];                        // stores the value received from the host in ASCCI text
          memcpy(argument, &receivedChars[13], 5); // copy characters from full input command 
          argument[5] = '\0';                      // termination character
          int newSampleRate = atoi(argument);
          helpers::printToLog("Received SetSampleRate Command: " + String(newSampleRate), verboseLevels::info);
          
          if(newSampleRate < minSampleRate || newSampleRate > maxSampleRate)
          {
            helpers::printToLog("Received Set Sample Rate command is below " + String(minSampleRate) + " or above " + String(maxSampleRate), verboseLevels::error);
          }
          else if(newSampleRate / downSampler < minSampleRate)
          {
            helpers::printToLog("Received Set Sample Rate command is too low in combination with configured downsampler: " + String(downSampler), verboseLevels::error);
          }
          else
          {     
            aquisitionTaskDuration = round(1000000 / newSampleRate);   // µs per cycle   //for aquisition of sensor data
            aquisitionTask.setInterval(aquisitionTaskDuration);
          }
      }
      memset(receivedChars, 0, sizeof(receivedChars) ); // clean command buffer
    }
    else if (strncmp(receivedChars, "SetDownSampleRate", 17) == 0) // command starting with "SetDownSampleRate", should always be followed by 4 integers, for example "SetDownSampleRate0016" for 16 times downsample
    {
      if (aquisitionTask.isEnabled())
      {
        helpers::printToLog("Received Set DownSample Rate command, but aquisition is active!", verboseLevels::error);
      }
      else
      {
          char argument[5];                        // stores the value received from the host in ASCCI text
          memcpy(argument, &receivedChars[17], 4); // copy characters from full input command 
          argument[4] = '\0';                      // termination character
          int newDownSampleRate = atoi(argument);
          helpers::printToLog("Received SetDownSampleRate Command: " + String(newDownSampleRate), verboseLevels::info);
          
          if(newDownSampleRate < 1 || newDownSampleRate > maxDownSampler)
          {
            helpers::printToLog("Received Set Down Sample Rate command is below 1 or above " + String(maxDownSampler), verboseLevels::error);
          }
          else if((1000000 / aquisitionTask.getInterval()) / newDownSampleRate < minSampleRate)
          {
            
            helpers::printToLog("Received Set Down Sample Rate command is too low in combination with configured sample rate: " + String(1000000 / aquisitionTask.getInterval()), verboseLevels::error);
          }
          else
          {     
            downSampler = newDownSampleRate; //set new down sample rate
          }
      }
      memset(receivedChars, 0, sizeof(receivedChars) ); // clean command buffer
    }

  // write data to Serial

  if (btBufferA.isBufferLocked() && btBufferB.isBufferLocked())
  {
    helpers::printToLog("at Send: both write buffers locked!", verboseLevels::error);
  }

  // There are two data buffers: A & B, in the code below control is taken on which buffer is send and which is locked.
  // When Buffer A is filled with more data than one sample, this buffer is locked, and send to Serial by the code below.
  // The code where the sensor data is put into the buffer (at the higher sample rate) now switches to the other buffer because the code below has locked the first buffer.
  // The second time the code below is ran, it will send the second buffer, as such causing a ping-pong sending mechanism.
  // The task that collects the sensor data has higher prio than the Serial sending data, allowing some jitter on the sending without hampering the sensor sample rate.

  if (btBufferA.isBufferLocked())
  {
    if (btBufferB.getPointerPosition() > bufferSampleSet.getBufferSize()) // Buffer B is available for sending
    {
      btBufferB.lockBuffer();   // lock buffer B before sending it to the serial Receiver
      btBufferA.unlockBuffer(); // make buffer A available for writing
      helpers::printToLog(" try to send buffer B of size (bytes): " + String(btBufferB.getPointerPosition()) + ", data: " + String(btBufferB.buffer[10]), verboseLevels::extraBuffers);
      if (Serial.availableForWrite())
      {
       Serial.write((uint8_t *)btBufferB.buffer, btBufferB.getPointerPosition()); // write buffer until pointer position
      }
      btBufferB.clearBuffer(); // clear buffer B & reset pointer
    }
  }
  else if (btBufferA.getPointerPosition() > bufferSampleSet.getBufferSize()) // Buffer A is available for sending
  {
    btBufferA.lockBuffer();   // lock buffer A before sending it to the serial Receiver
    btBufferB.unlockBuffer(); // make buffer B available for writing

    helpers::printToLog(" try to send buffer A of size (bytes): " + String(btBufferA.getPointerPosition()) + ", data: " + String(btBufferA.buffer[10]), verboseLevels::extraBuffers);
    if (Serial.availableForWrite())
    {
      Serial.write((uint8_t *)btBufferA.buffer, btBufferA.getPointerPosition()); // write buffer until pointer position
    }
    btBufferA.clearBuffer(); // clear buffer A & reset pointer
  }

} // end communication Task

void setup()
{
  Serial.begin(serialBaudrate); // Start Serial device name
  //  Serial.setTimeout(serialTimeOut); //set read timeout
  delay(1000);
  // Serial.println("Version: " + (String)versionNumber + " Compiled at " + (String)__DATE__ + " " + (String)__TIME__);
  helpers::printToLog("Version: " + (String)versionNumber + " Compiled at " + (String)__DATE__ + " " + (String)__TIME__, verboseLevels::info);

  // set SPI for ADC
  SPI.setMOSI(MOSIpin);
  SPI.setMISO(MISOpin);
  SPI.setSCK(SCKpin);
  // set the chipSelectADCPin as an output:
  pinMode(chipSelectADCPin, OUTPUT);
  digitalWrite(chipSelectADCPin, HIGH);
  // start the SPI library:
  SPI.begin();
  // Set clock divider to 42
  // SPI.setClockDivider(spiClockDivider); //seems to be ignored by the Teensy.

  // initialize the buffers used to store data

  bufferMillis.initBuffer(4, 0); // microtime (4 bytes, signed integer): 00 01 01 01   big-e
  bufferADC.initBuffer(16, 0);   // ADC floats (4 bytes each)
  // bufferBatt.initBuffer(16, 0);  // ADC floats (4 bytes each)

  // bufferCRC.initBuffer(2, 0);   // CRC-16 data buffer, 2x2 byte
  bufferSampleSet.initBuffer(startBytesSize + bufferMillis.getBufferSize() + bufferADC.getBufferSize(), 0); //+ bufferBatt.getBufferSize()  // complete buffer for a single sample set
  btBufferA.initBuffer(communicationBufferSize, buffers::bufferSampleSet.getBufferSize());                  // start pointer at sizeof(buffer) (otherwise you miss the first sample set)
  btBufferB.initBuffer(communicationBufferSize, buffers::bufferSampleSet.getBufferSize());                  // start pointer at sizeof(buffer) (otherwise you miss the first sample set)

  // start the scheduler worker. Tasks can be added the scheduler and given a certain execution frequency.
  runner.init();

  runner.addTask(aquisitionTask);
  helpers::printToLog("prepaired aquisition task. Waiting for 0.5 seconds.", verboseLevels::info);
  delay(500);

  runner.addTask(communicationTask);
  helpers::printToLog("prepaired communication task. Waiting for 0.5 seconds.", verboseLevels::info);
  delay(500);

  communicationTask.setSchedulingOption(TASK_SCHEDULE_NC); // disable catch-ups
  aquisitionTask.setSchedulingOption(TASK_SCHEDULE_NC);    // disable catch-ups

  communicationTask.enable();
  helpers::printToLog("Communication task started!", verboseLevels::info);

  //aquisitionTask.enable();
  //helpers::printToLog("enabled aquisitionTask (start sampling)", verboseLevels::info);
}
void loop()
{
  // trigger the task scheduler, all other actions belong to one of the two tasks.
  runner.execute();
}